"""Pydise - Detector."""
import argparse
import ast
import copy
import os
import logging
import linecache
from glob import glob

# TODO : An ast.Expr may not generate a side effect, but it's hard to distinguish, so
#        by default an ast.Expr will be defined as a possible side-effect generator.
PATTERNS_SIDE_EFFECTS = (ast.Expr, ast.Raise, ast.Assert, ast.Delete)
PATTERNS_IGNORED = ["# no-pydise", "# no_pydise"]
dict_assign = dict()
dict_functions = dict()


class PydiseSideEffects(Exception):
    """Pydise Exception."""

    def __init__(self, message=""):
        """Init exception message."""
        self.message = message
        super().__init__(self.message)


class PydiseLoadError(Exception):
    """Pydise Exception."""

    def __init__(self, message=""):
        """Init exception message."""
        self.message = message
        super().__init__(self.message)


class RewriteName(ast.NodeTransformer):
    """Class for Rewrite some nodes in an AST Tree."""

    def visit_Name(self, node):
        """Replace ast.Name to an ast value."""
        return dict_assign.get(node.id, node)


class PyDiseLoader(object):
    """Class used to define strategy to load a data and return an ast_module."""

    def __init__(
        self,
        filename=None,
        file=None,
        ast_tree=None,
    ):
        """Init data to load, based on filename / file or an ast_tree."""
        if filename:
            self.filename = filename
            self.load_from_filename(filename)
        elif file:
            self.filename = None
            self.load_from_file(file)
        elif ast_tree:
            self.filename = None
            self.load_from_ast(ast_tree)
        else:
            self.filename = None
            self.ast_module = None

    def load_from_filename(self, filename):
        """Load a file from a filename, and set an ast_module."""
        self.filename = filename
        if not os.path.isfile(self.filename):
            print("'{}' isn't a file.")
        with open(self.filename, "rb") as file:
            self.ast_module = ast.parse(file.read(), filename=self.filename)

    def load_from_file(self, file):
        """Load a file, and set an ast_module."""
        self.ast_module = ast.parse(file.read())

    def load_from_ast(self, ast_tree):
        """Load an ast module."""
        self.ast_module = ast_tree


class PyDise(object):
    """Main class."""

    def __init__(
        self,
        filename=None,
        file=None,
        ast_tree=None,
        patterns_ignored=None,
        on_error="logger",
    ):
        """Init."""
        try:
            pydise_loader_obj = PyDiseLoader(
                filename=filename, file=file, ast_tree=ast_tree
            )
            self.filename = pydise_loader_obj.filename
            self.ast_module = pydise_loader_obj.ast_module
        except SyntaxError:
            raise PydiseLoadError
        self.on_error = on_error
        self.patterns_ignored = copy.copy(PATTERNS_IGNORED)
        if isinstance(patterns_ignored, list):
            self.patterns_ignored.extend(patterns_ignored)

        self.side_effects = {"warnings": list(), "errors": list()}

    def save_variables(self, ast_assign):
        """Save variables into a dictionnary."""
        if not isinstance(ast_assign, ast.Assign):
            logging.error("Not an AST Assign.")

        for target in ast_assign.targets:
            if hasattr(target, "id"):
                dict_assign[target.id] = ast_assign.value

    def save_functions(self, ast_def):
        """Save functions / class to a dictionnary."""
        # TODO : Improve this method to retrieve sub function
        if not isinstance(ast_def, (ast.FunctionDef, ast.ClassDef)):
            logging.error("Not an AST FunctionDef or ClassDef.")
        dict_functions[ast_def.name] = ast_def

    def _notify(self, tree_element, level=logging.ERROR, on_error=None):
        """Notifying assertion."""
        raw_line = linecache.getline(
            self.filename, tree_element.lineno, module_globals=None
        )
        message = f"{self.filename}:{tree_element.lineno} -> Side effects detected : {raw_line}"

        if on_error == "logger":
            logging.log(level, message)
        elif on_error == "raise":
            raise PydiseSideEffects(message)
        else:
            pass

    def notify(self, on_error=None):
        """Use to notify user."""
        if on_error is None:
            on_error = self.on_error
        side_effects_warnings = list(set(self.side_effects.get("warnings", list())))
        side_effects_errors = list(set(self.side_effects.get("errors", list())))

        for side_effect in side_effects_warnings:
            self._notify(side_effect, level=logging.WARNING, on_error=on_error)

        for side_effect in side_effects_errors:
            self._notify(side_effect, level=logging.ERROR, on_error=on_error)

    def analyze(self, ast_module=None):
        """Analyze the AST Module.

        Return a dict with errors of the AST nodes that will be executed during import.
        """
        if ast_module is None:
            ast_module = self.ast_module
        if not isinstance(ast_module, ast.Module):
            logging.error("Not an AST Module.")
        else:
            tree_elements = ast_module.body

            for tree_element in tree_elements:
                tree_element = RewriteName().visit(tree_element)
                self.get_side_effects(tree_element)
        return self.side_effects

    def is_side_effects(self, node):
        """Check if the ast node could generate a side effect."""
        if isinstance(node, PATTERNS_SIDE_EFFECTS):
            # When ast.Expr(value=ast.Constant) assuming it's a docstring -> Ignored
            if hasattr(node, "value") and isinstance(node.value, ast.Constant):
                return False

            # Exclusion based on a line pattern
            raw_line = linecache.getline(
                self.filename, node.lineno, module_globals=None
            )
            for pattern_ignored in self.patterns_ignored:
                if pattern_ignored in raw_line:
                    return False
            return True

    def get_side_effects(self, tree_element, recursive=False):
        """Recursively dig into an ast tree and found side effects."""
        # TODO : Add "try/finally"
        # TODO : Add "match"

        if self.is_side_effects(tree_element):
            self.side_effects["errors"].append(tree_element)

        # Save the first level functions / class.
        if isinstance(tree_element, (ast.FunctionDef, ast.ClassDef)):
            self.save_functions(tree_element)

        if isinstance(tree_element, ast.FunctionDef):
            self.get_side_effects(tree_element.args)

        if isinstance(tree_element, ast.arguments):
            for default in tree_element.defaults:
                self.get_side_effects(default)

            for kw_default in tree_element.kw_defaults:
                self.get_side_effects(kw_default)

        if recursive:
            # For several type of statement like for/while/try/if/with/raise etc...
            if hasattr(tree_element, "body"):
                for sub_element in tree_element.body:
                    self.get_side_effects(sub_element)

            # For object Assign()
            if hasattr(tree_element, "targets"):
                function_name = tree_element.value.func.id
                if dict_functions.get(function_name):
                    self.get_side_effects(dict_functions.get(function_name))
        else:
            if isinstance(tree_element, ast.Call):
                if hasattr(tree_element, "func") and hasattr(tree_element.func, "id"):
                    dest_call = dict_functions.get(tree_element.func.id)
                    if isinstance(dest_call, ast.ClassDef):
                        for body_data in dest_call.body:
                            if (
                                hasattr(body_data, "name")
                                and body_data.name == "__init__"
                            ):
                                self.get_side_effects(body_data, recursive=True)
                    else:
                        self.get_side_effects(dest_call, recursive=True)

            if isinstance(tree_element, ast.Assign):
                for child in list(ast.iter_child_nodes(tree_element)):
                    # catch dynamic assignment like "a = foo()" and check if this functions could generate a side effects
                    if isinstance(child, ast.Call):
                        self.get_side_effects(child)
                else:
                    self.save_variables(tree_element)

            # if isinstance(tree_element, (ast.ClassDef, ast.Try, ast.With)):
            if isinstance(tree_element, (ast.Try, ast.With)):
                self.get_side_effects(tree_element, recursive=True)

            if isinstance(tree_element, ast.For):
                loop_var = ast.unparse(tree_element.target)
                loop_iter = ast.unparse(tree_element.iter)

                try:
                    eval_code = eval(f"[True for {loop_var} in {loop_iter}]")
                    eval_code = True if True in eval_code else False
                except Exception:
                    eval_code = False

                if eval_code:
                    self.get_side_effects(tree_element, recursive=True)
                    is_break = (
                        True
                        if True in [isinstance(x, ast.Break) for x in tree_element.body]
                        else False
                    )

                    if not is_break and tree_element.orelse:
                        for sub_condition in tree_element.orelse:
                            self.get_side_effects(sub_condition)
                else:
                    if tree_element.orelse:
                        for sub_condition in tree_element.orelse:
                            self.get_side_effects(sub_condition)

            if isinstance(tree_element, (ast.If, ast.While)):
                # Skip test when it's a function / object
                if isinstance(tree_element.test, ast.Call):
                    # self.side_effects["warnings"].append(f"Possible side-effect - {tree_element.lineno}")
                    self.side_effects["warnings"].append(tree_element.test)
                else:
                    unparse_code = ast.unparse(tree_element.test)
                    try:
                        eval_code = eval(unparse_code)
                    except Exception:
                        eval_code = False

                    if "__main__" not in unparse_code:
                        if eval_code:
                            self.get_side_effects(tree_element, recursive=True)
                        else:
                            if tree_element.orelse:
                                for sub_condition in tree_element.orelse:
                                    self.get_side_effects(sub_condition)

        return self.side_effects


def get_filenames(args):
    """Return a list of files based on args or by default, from the current directory."""
    list_files = list()
    if os.path.isdir(args.filename):
        for root, sub_directories, files in os.walk(args.filename):
            for filename in files:
                if str(filename).endswith(".py"):
                    list_files.append(os.path.join(root, filename))
    elif os.path.isfile(args.filename) and str(args.filename).endswith(".py"):
        list_files.append(args.filename)
    else:
        for file in glob(os.path.join(os.path.curdir, f"{args.filename}*.py")):
            list_files.append(file)
    return list_files


def main(filename, on_error="logger", pattern_ignored=None):
    """Script main entry point."""
    try:
        pydise_object = PyDise(
            filename=filename, on_error=on_error, patterns_ignored=pattern_ignored
        )
        pydise_object.analyze()
        pydise_object.notify()
        return pydise_object.side_effects
    except PydiseLoadError:
        print(f"!!! - {filename} -> unable to read the file (maybe python2 ?)")


def run():
    """Run."""
    parser = argparse.ArgumentParser()
    # parser.add_argument("--filename", help="file to check")
    parser.add_argument(
        "filename", help="file to check (wildcard)", type=str, nargs="?", default="."
    )
    parser.add_argument(
        "--list-only",
        help="list the detected files without checking errors.",
        action="store_true",
    )
    parser.add_argument(
        "--pattern-ignored",
        help="ignore line containing the pattern, multiple patterns can be setted, "
        "the pattern is added to the default patterns.",
        action="append",
    )
    args = parser.parse_args()

    list_files = get_filenames(args)

    if args.list_only:
        print("Detected files : ")
        for file in list_files:
            print(f"* {file}")
        exit(0)

    for path in list_files:
        main(filename=path, pattern_ignored=args.pattern_ignored)


if __name__ == "__main__":
    run()



from io import StringIO
import pytest
import pydise.detector
import jinja2


template_for = jinja2.Template(
    """
for {{ loop }}:
    print("foo")
"""
)
template_for_else = jinja2.Template(
    """
for {{ loop }}:
    {{ break_pass }}
else:
    print("foo")
"""
)

list_loop_true = [
    "i in range(5)",
    "i in ['foo']",
    "i in {'foo': 'bar'}.items()",
]

list_loop_false = [
    "i in range(0)",
    "i in []",
    "i in {}",
]

expected_for = ":3 -> Side effects detected :"
expected_for_else = ":5 -> Side effects detected :"


@pytest.mark.parametrize("loop_true", list_loop_true)
def test_for_else_iter_ok_no_break(loop_true):
    code_test = jinja2.Template.render(
        template_for_else, loop=loop_true, break_pass="pass"
    )

    test_parse = StringIO(code_test)

    with pytest.raises(pydise.detector.PydiseSideEffects) as exc_info:
        pydise_object = pydise.detector.PyDise(file=test_parse)
        pydise_object.analyze()
        pydise_object.notify(on_error="raise")

    assert expected_for_else in str(exc_info.value)


@pytest.mark.parametrize("loop_true", list_loop_true)
def test_for_else_iter_ok_break(loop_true):
    code_test = jinja2.Template.render(
        template_for_else, loop=loop_true, break_pass="break"
    )

    test_parse = StringIO(code_test)

    pydise_object = pydise.detector.PyDise(file=test_parse)
    pydise_object.analyze()
    pydise_object.notify(on_error="raise")


@pytest.mark.parametrize("loop_false", list_loop_false)
def test_for_else_iter_ko(loop_false):
    code_test = jinja2.Template.render(
        template_for_else, loop=loop_false, break_pass="break"
    )

    test_parse = StringIO(code_test)

    with pytest.raises(pydise.detector.PydiseSideEffects) as exc_info:
        pydise_object = pydise.detector.PyDise(file=test_parse)
        pydise_object.analyze()
        pydise_object.notify(on_error="raise")

    assert expected_for_else in str(exc_info.value)


@pytest.mark.parametrize("loop_true", list_loop_true)
def test_for_iter_ok(loop_true):
    code_test = jinja2.Template.render(template_for, loop=loop_true)

    test_parse = StringIO(code_test)

    with pytest.raises(pydise.detector.PydiseSideEffects) as exc_info:
        pydise_object = pydise.detector.PyDise(file=test_parse)
        pydise_object.analyze()
        pydise_object.notify(on_error="raise")

    assert expected_for in str(exc_info.value)


@pytest.mark.parametrize("loop_false", list_loop_false)
def test_for_iter_ko(loop_false):
    code_test = jinja2.Template.render(template_for, loop=loop_false)

    test_parse = StringIO(code_test)

    pydise_object = pydise.detector.PyDise(file=test_parse)
    pydise_object.analyze()
    pydise_object.notify(on_error="raise")


def test_try_ko():
    code_test = """
try:
    print("foo")
except Exception:
    pass
"""

    test_parse = StringIO(code_test)

    with pytest.raises(pydise.detector.PydiseSideEffects) as exc_info:
        pydise_object = pydise.detector.PyDise(file=test_parse)
        pydise_object.analyze()
        pydise_object.notify(on_error="raise")

    assert ":3 -> Side effects detected" in str(exc_info.value)


def test_try_ok():
    code_test = """
try:
    pass
except Exception:
    print("foo")
"""

    test_parse = StringIO(code_test)

    pydise_object = pydise.detector.PyDise(file=test_parse)
    pydise_object.analyze()
    pydise_object.notify(on_error="raise")


@pytest.mark.skip(reason="Not Implemented")
def test_except_ko():
    code_test = """
try:
    a = a
except NameError:
    print("foo")
"""

    test_parse = StringIO(code_test)

    with pytest.raises(pydise.detector.PydiseSideEffects) as exc_info:
        pydise_object = pydise.detector.PyDise(file=test_parse)
        pydise_object.analyze()
        pydise_object.notify(on_error="raise")

    assert ":3 -> Side effects detected" in str(exc_info.value)


def test_except_ok():
    code_test = """
try:
    a = a
except NameError:
    pass
"""

    test_parse = StringIO(code_test)

    pydise_object = pydise.detector.PyDise(file=test_parse)
    pydise_object.analyze()
    pydise_object.notify(on_error="raise")



from io import StringIO
import pytest
import pydise.detector
import jinja2

template_while = jinja2.Template(
    """
while {{ condition }}:
    print("foo")
"""
)
template_while_else = jinja2.Template(
    """
while {{ condition }}:
    pass
else:
    print("foo")
"""
)

template_if = jinja2.Template(
    """
if {{ condition }}:
    print("foo")
"""
)
expected_if = ":3 -> Side effects detected :"
expected_while = ":3 -> Side effects detected :"
expected_while_else = ":5 -> Side effects detected :"

list_conditions_true = [
    "True",
    "False or True",
    "True and True",
    "True == True",
    "True != False",
    '"1"',
    '"1" or ""',
    '"1" and "1"',
    '"1" == "1"',
    '"1" != ""',
    "1",
    "1 or 0",
    "1 and 1",
    "1 == 1",
    "1 != 0",
    '"a" in ["a"]',
    '"a" not in ["b"]',
    "1 in [1]",
    "1 not in [0]",
    '["a", "b"] == ["a", "b"]',
    '["a", "b"] != ["b", "a"]',
    "True and (True == True) and 1==1",
    "True and (True == True) or 2==1",
    "(True and (True != True)) or 1==1",
]

list_conditions_false = [
    "False",
    "False and True",
    "False or False",
    "True == False",
    "False != False",
    '""',
    '"" or ""',
    '"1" and ""',
    '"1" == ""',
    '"1" != "1"',
    "0",
    "0 or 0",
    "1 and 0",
    "1 == 2",
    "1 != 1",
    '"a" in ["b"]',
    '"a" not in ["a"]',
    "1 in []",
    "1 not in [1]",
    '["a", "b"] != ["a", "b"]',
    '["a", "b"] == ["b", "a"]',
    "True and (True == True) and 2==1",
    "True and (True == False) or 2==1",
    "(True and (True != True)) or 2==1",
]


@pytest.mark.parametrize("condition_true", list_conditions_true)
def test_if_ko(condition_true):
    code_test = jinja2.Template.render(template_if, condition=condition_true)

    test_parse = StringIO(code_test)

    with pytest.raises(pydise.detector.PydiseSideEffects) as exc_info:
        pydise_object = pydise.detector.PyDise(file=test_parse)
        pydise_object.analyze()
        pydise_object.notify(on_error="raise")

    assert expected_if in str(exc_info.value)


@pytest.mark.parametrize("condition_false", list_conditions_false)
def test_if_ok(condition_false):
    code_test = jinja2.Template.render(template_if, condition=condition_false)

    test_parse = StringIO(code_test)

    pydise_object = pydise.detector.PyDise(file=test_parse)
    pydise_object.analyze()
    pydise_object.notify(on_error="raise")


@pytest.mark.parametrize("condition_true", list_conditions_true)
def test_while_iter_ok(condition_true):
    code_test = jinja2.Template.render(template_while, condition=condition_true)

    test_parse = StringIO(code_test)

    with pytest.raises(pydise.detector.PydiseSideEffects) as exc_info:
        pydise_object = pydise.detector.PyDise(file=test_parse)
        pydise_object.analyze()
        pydise_object.notify(on_error="raise")

    assert expected_while in str(exc_info.value)


@pytest.mark.parametrize("condition_false", list_conditions_false)
def test_while_iter_ko(condition_false):
    code_test = jinja2.Template.render(template_while, condition=condition_false)

    test_parse = StringIO(code_test)

    pydise_object = pydise.detector.PyDise(file=test_parse)
    pydise_object.analyze()
    pydise_object.notify(on_error="raise")


@pytest.mark.parametrize("condition_false", list_conditions_false)
def test_while_else_iter_ko(condition_false):
    code_test = jinja2.Template.render(template_while_else, condition=condition_false)

    test_parse = StringIO(code_test)

    with pytest.raises(pydise.detector.PydiseSideEffects) as exc_info:
        pydise_object = pydise.detector.PyDise(file=test_parse)
        pydise_object.analyze()
        pydise_object.notify(on_error="raise")

    assert expected_while_else in str(exc_info.value)



from io import StringIO
import pytest
import pydise.detector

expected_error = ":1 -> Side effects detected :"
list_ko = [
    "print('foo')",
    "assert('foo')",
    "time.sleep(1)",
    "exit(0)",
    "del x",
    "a = lambda x: print(x); a('foo')",
]

list_ok = ["False", "a = 3", "a = lambda x: print(x)", "pass"]


@pytest.mark.parametrize("statement_ko", list_ko)
def test_base_ko(statement_ko):
    test_parse = StringIO(statement_ko)

    with pytest.raises(pydise.detector.PydiseSideEffects) as exc_info:
        pydise_object = pydise.detector.PyDise(file=test_parse)
        pydise_object.analyze()
        pydise_object.notify(on_error="raise")

    assert expected_error in str(exc_info.value)


@pytest.mark.parametrize("statement_ok", list_ok)
def test_base_ok(statement_ok):
    test_parse = StringIO(statement_ok)

    pydise_object = pydise.detector.PyDise(file=test_parse)
    pydise_object.analyze()
    pydise_object.notify(on_error="raise")



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
"""Packaging Setup."""
from setuptools import setup
import os
import datetime

if __name__ == "__main__":
    try:
        setup(
            use_scm_version=True,
            setup_requires=["setuptools_scm"],
        )
    except Exception:
        now = datetime.datetime.now()
        date_as_version = str(now.strftime("%Y.%m.%d.%H.%M.%S"))

        version = os.environ.get("CI_COMMIT_TAG")
        version = version if version else date_as_version
        setup(version=version)
